pub mod app;
pub mod cli;
pub mod comment_vim;
pub mod config;
pub mod editor;
pub mod error;
pub mod forge;
pub mod handler;
pub mod hash;
pub mod input;
pub mod model;
pub mod output;
pub mod persistence;
pub mod process;
pub mod profile;
pub mod review_cli;
pub mod review_store;
pub mod slug;
pub mod syntax;
pub mod terminal_state;
pub mod text_edit;
pub mod theme;
pub mod tuicrignore;
pub mod ui;
pub mod update;
pub mod vcs;

pub use error::{Result, TuicrError};
pub use model::{Comment, CommentType, LineRange, LineSide, ReviewSession, SessionDiffSource};
pub use review_store::{
    AddCommentRequest, CommentTarget, ReviewStore, SessionRef, SessionSummary,
    add_comment_to_session,
};
